// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
// ** Copyright UCAR (c) 1992 - 2015
// ** University Corporation for Atmospheric Research (UCAR)
// ** National Center for Atmospheric Research (NCAR)
// ** Research Applications Lab (RAL)
// ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*



////////////////////////////////////////////////////////////////////////


using namespace std;

#include <iostream>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <cstdio>
#include <cmath>

#include "nc_utils.h"
#include "nc_grid.h"


////////////////////////////////////////////////////////////////////////


static bool read_nc_st_grid     (NcFile &, Grid &);
static bool read_nc_lc_grid     (NcFile &, Grid &);
static bool read_nc_latlon_grid (NcFile &, Grid &);

static void write_nc_st_grid     (NcFile &, const StereographicData &);
static void write_nc_lc_grid     (NcFile &, const LambertData &);
static void write_nc_latlon_grid (NcFile &, const LatLonData &);


////////////////////////////////////////////////////////////////////////


bool read_nc_grid(NcFile & f, Grid & g)

{

bool status = false;
const ConcatString proj = string_att(f, "Projection");


g.clear();


if ( proj == "Polar Stereographic" )  {

   status = read_nc_st_grid(f, g);

} else if ( proj == "Lambert Conformal" )  {

   status = read_nc_lc_grid(f, g);

} else if ( proj == "LatLon" )  {

   status = read_nc_latlon_grid(f, g);

} else {

   mlog << Error << "\n\n  read_nc_grid() -> haven't written code to parse \"" << proj << "\" grids yet!\n\n";

   return ( false );

}


   //
   //  done
   //

return ( status );

}


////////////////////////////////////////////////////////////////////////


bool write_nc_grid(NcFile & f, const Grid & g)

{

GridInfo info = g.info();

if ( !(info.ok()) )  {

   mlog << Error << "\n\n  write_nc_grid(NcFile &, const Grid &) -> can't get information from grid!\n\n";

   exit ( 1 );

}

     if ( info.st )  write_nc_st_grid     (f, *(info.st));
else if ( info.lc )  write_nc_lc_grid     (f, *(info.lc));
else if ( info.ll )  write_nc_latlon_grid (f, *(info.ll));
else {

   mlog << Error << "\n\n  bool write_nc_grid(NcFile &, const Grid &) -> unsupported projection type\n\n";

   exit ( 1 );

}

   //
   //  done
   //

return ( false );

}


////////////////////////////////////////////////////////////////////////


bool read_nc_st_grid(NcFile & f, Grid & g)

{

StereographicData data;
const char * c = (const char *) 0;

   //
   //  name
   //

data.name = "Unknown stereographic";

   //
   //  hemisphere
   //

c = string_att(f, "hemisphere");

data.hemisphere = c[0];

   //
   //  scale latitude
   //

data.scale_lat = string_att_as_double(f, "scale_lat");

   //
   //  lat/lon pin
   //

data.lat_pin = string_att_as_double(f, "lat_pin");
data.lon_pin = string_att_as_double(f, "lon_pin");

data.lon_pin *= -1.0;

   //
   //  x/y pin
   //

data.x_pin = string_att_as_double(f, "x_pin");
data.y_pin = string_att_as_double(f, "y_pin");

   //
   //  orientation longitude
   //

data.lon_orient = string_att_as_double(f, "lon_orient");

data.lon_orient *= -1.0;

   //
   //  D, R
   //

data.d_km = string_att_as_double(f, "d_km");
data.r_km = string_att_as_double(f, "r_km");

   //
   //  Nx, Ny
   //

data.nx = string_att_as_int(f, "nx");
data.ny = string_att_as_int(f, "ny");


   //
   //  done
   //

g.set(data);

return ( true );

}


////////////////////////////////////////////////////////////////////////


bool read_nc_lc_grid(NcFile & f, Grid & g)

{

LambertData data;
g.clear();

   //
   //  name
   //

data.name = "Unknown lambert";

   //
   //  scale latitudes
   //

data.scale_lat_1 = string_att_as_double(f, "scale_lat_1");
data.scale_lat_2 = string_att_as_double(f, "scale_lat_2");

   //
   //  lat/lon pin
   //

data.lat_pin = string_att_as_double(f, "lat_pin");
data.lon_pin = string_att_as_double(f, "lon_pin");

data.lon_pin *= -1.0;

   //
   //  x/y pin
   //

data.x_pin = string_att_as_double(f, "x_pin");
data.y_pin = string_att_as_double(f, "y_pin");

   //
   //  orientation longitude
   //

data.lon_orient = string_att_as_double(f, "lon_orient");

data.lon_orient *= -1.0;

   //
   //  D, R
   //

data.d_km = string_att_as_double(f, "d_km");
data.r_km = string_att_as_double(f, "r_km");

   //
   //  Nx, Ny
   //

data.nx = string_att_as_int(f, "nx");
data.ny = string_att_as_int(f, "ny");

   //
   //  done
   //

g.set(data);

return ( true );

}


////////////////////////////////////////////////////////////////////////


bool read_nc_latlon_grid(NcFile & f, Grid & g)

{

LatLonData data;

g.clear();

   //
   //  name
   //

data.name = "Unknown latlon";

   //
   //  lower-left lat/lon
   //

data.lat_ll = string_att_as_double(f, "lat_ll");
data.lon_ll = string_att_as_double(f, "lon_ll");

data.lon_ll = -(data.lon_ll);

   //
   //  lat/lon deltas
   //

data.delta_lat = string_att_as_double(f, "delta_lat");
data.delta_lon = string_att_as_double(f, "delta_lon");

   //
   //  grid size
   //

data.Nlat = string_att_as_int(f, "Nlat");
data.Nlon = string_att_as_int(f, "Nlon");


   //
   //  done
   //

g.set(data);

return ( true );

}


////////////////////////////////////////////////////////////////////////


void write_nc_st_grid(NcFile & f, const StereographicData & data)

{

char junk[256];
char j2[256];

   //
   //  name
   //

f.add_att("Projection", "Polar Stereographic");

   //
   //  hemisphere
   //

junk[0] = data.hemisphere;
junk[1] = (char) 0;

f.add_att("hemisphere", junk);

   //
   //  scale latitude
   //

sprintf(j2, "%.5f", data.scale_lat);

fix_float(j2);

sprintf(junk, "%s degrees_north", j2);

f.add_att("scale_lat", junk);

   //
   //  lat/lon pin point
   //

sprintf(junk, "%.5f", data.lat_pin);

fix_float(junk);

f.add_att("lat_pin", junk);


sprintf(junk, "%.5f", -(data.lon_pin));

fix_float(junk);

f.add_att("lon_pin", junk);

   //
   //  x/y pin point
   //

sprintf(junk, "%.5f", data.x_pin);

fix_float(junk);

f.add_att("x_pin", junk);


sprintf(junk, "%.5f", data.y_pin);

fix_float(junk);

f.add_att("y_pin", junk);

   //
   //  orientation longitude
   //

sprintf(junk, "%.5f", -(data.lon_orient));

fix_float(junk);

f.add_att("lon_orient", junk);

   //
   //  D and R
   //

sprintf(j2, "%.5f", data.d_km);

fix_float(j2);

sprintf(junk, "%s km", j2);

f.add_att("d_km", junk);


sprintf(j2, "%.5f", data.r_km);

fix_float(j2);

sprintf(junk, "%s km", j2);

f.add_att("r_km", junk);

   //
   //  nx and ny
   //

sprintf(junk, "%d", data.nx);

f.add_att("nx", junk);


sprintf(junk, "%d", data.ny);

f.add_att("ny", junk);


   //
   //  done
   //

return;

}


////////////////////////////////////////////////////////////////////////


void write_nc_lc_grid(NcFile & f, const LambertData & data)

{

char junk[256];
char j2[256];

   //
   //  name
   //

f.add_att("Projection", "Lambert Conformal");

   //
   //  scale latitudes
   //

sprintf(junk, "%.5f", data.scale_lat_1);

fix_float(junk);

f.add_att("scale_lat_1", junk);


sprintf(junk, "%.5f", data.scale_lat_2);

fix_float(junk);

f.add_att("scale_lat_2", junk);

   //
   //  lat/lon pin point
   //

sprintf(junk, "%.5f", data.lat_pin);

fix_float(junk);

f.add_att("lat_pin", junk);


sprintf(junk, "%.5f", -(data.lon_pin));

fix_float(junk);

f.add_att("lon_pin", junk);

   //
   //  x/y pin point
   //

sprintf(junk, "%.5f", data.x_pin);

fix_float(junk);

f.add_att("x_pin", junk);


sprintf(junk, "%.5f", data.y_pin);

fix_float(junk);

f.add_att("y_pin", junk);

   //
   //  orientation longitude
   //

sprintf(junk, "%.5f", -(data.lon_orient));

fix_float(junk);

f.add_att("lon_orient", junk);

   //
   //  D and R
   //

sprintf(j2, "%.5f", data.d_km);

fix_float(j2);

sprintf(junk, "%s km", j2);

f.add_att("d_km", junk);


sprintf(j2, "%.5f", data.r_km);

fix_float(j2);

sprintf(junk, "%s km", j2);

f.add_att("r_km", junk);

   //
   //  nx and ny
   //

sprintf(junk, "%d", data.nx);

f.add_att("nx", junk);


sprintf(junk, "%d", data.ny);

f.add_att("ny", junk);


   //
   //  done
   //

return;

}


////////////////////////////////////////////////////////////////////////


void write_nc_latlon_grid (NcFile & f, const LatLonData & data)

{

char junk[256];


   //
   //  name
   //

f.add_att("Projection", "LatLon");

   //
   //  lower left point
   //

sprintf(junk, "%.5f", data.lat_ll);

fix_float(junk);

f.add_att("lat_ll", junk);


sprintf(junk, "%.5f", -(data.lon_ll));

fix_float(junk);

f.add_att("lon_ll", junk);

   //
   //  lat/lon deltas
   //

sprintf(junk, "%.5f", data.delta_lat);

fix_float(junk);

f.add_att("delta_lat", junk);


sprintf(junk, "%.5f", data.delta_lon);

fix_float(junk);

f.add_att("delta_lon", junk);

   //
   //  grid size
   //

sprintf(junk, "%d", data.Nlat);

fix_float(junk);

f.add_att("Nlat", junk);


sprintf(junk, "%d", data.Nlon);

fix_float(junk);

f.add_att("Nlon", junk);


   //
   //  done
   //

return;

}


////////////////////////////////////////////////////////////////////////


